import typing
from dataclasses import dataclass

from flexparser import flexparser as fp


@dataclass(frozen=True)
class NotAValidIdentifier(fp.ParsingError):

    value: str


@dataclass(frozen=True)
class CannotParseToFloat(fp.ParsingError):

    value: str


@dataclass(frozen=True)
class Open(fp.ParsedStatement):
    @classmethod
    def from_string(cls, s: str):
        if s == "@begin":
            return cls()
        return None


@dataclass(frozen=True)
class Close(fp.ParsedStatement):
    @classmethod
    def from_string(cls, s: str):
        if s == "@end":
            return cls()
        return None


@dataclass(frozen=True)
class Comment(fp.ParsedStatement):

    s: str

    @classmethod
    def from_string(cls, s: str):
        if s.startswith("#"):
            return cls(s)
        return None


@dataclass(frozen=True)
class EqualFloat(fp.ParsedStatement):

    a: str
    b: float

    @classmethod
    def from_string(cls, s: str) -> fp.FromString["EqualFloat"]:
        if "=" not in s:
            return None

        a, b = s.split("=")
        a = a.strip()
        b = b.strip()

        if not str.isidentifier(a):
            return NotAValidIdentifier(a)

        try:
            b = float(b)
        except Exception:
            return CannotParseToFloat(b)

        return cls(a, b)


class MyBlock(fp.Block):

    opening: fp.Single[Open]
    body: fp.Multi[typing.Union[Comment, EqualFloat]]
    closing: fp.Single[Close]


class MyRoot(fp.RootBlock):

    body: fp.Multi[typing.Union[Comment, EqualFloat]]


class MyParser(fp.Parser):

    _root_block_class = MyRoot


class MyRootWithBlock(fp.RootBlock):

    body: fp.Multi[typing.Union[Comment, EqualFloat, MyBlock]]


class MyParserWithBlock(fp.Parser):

    _root_block_class = MyRootWithBlock



import pytest

from flexparser import flexparser as fp
from flexparser.testsuite.common import (
    CannotParseToFloat,
    Close,
    Comment,
    EqualFloat,
    MyBlock,
    MyRoot,
    NotAValidIdentifier,
    Open,
)

MyBlock2 = fp.Block.subclass_with(
    opening=Open, body=(Comment, EqualFloat), closing=Close
)

MyRoot2 = fp.RootBlock.subclass_with(body=(Comment, EqualFloat))


def test_formatting():
    obj = EqualFloat.from_string("a = 3.1")
    assert obj.format_line_col == "(line: -1, col: -1)"
    obj.set_line_col(10, 3)
    assert obj.format_line_col == "(line: 10, col: 3)"
    assert str(obj) == "EqualFloat(lineno=10, colno=3, a='a', b=3.1)"

    obj = EqualFloat.from_string("%a = 3.1")
    assert obj.origin == ""
    assert obj.format_line_col == "(line: -1, col: -1)"
    obj.set_line_col(10, 3)
    assert obj.format_line_col == "(line: 10, col: 3)"

    obj1 = obj.copy_with(("pack", "nam"))
    assert obj1.origin_ == "resource (package: pack, name: nam)"

    obj2 = obj.copy_with("/util/bla.txt")
    assert obj2.origin_ == "/util/bla.txt"


def test_parse_equal_float():
    assert EqualFloat.from_string("a = 3.1") == EqualFloat("a", 3.1)
    assert EqualFloat.from_string("a") is None

    assert EqualFloat.from_string("%a = 3.1") == NotAValidIdentifier("%a")
    assert EqualFloat.from_string("a = 3f1") == CannotParseToFloat("3f1")

    obj = EqualFloat.from_string("a = 3f1")
    assert str(obj) == "CannotParseToFloat(lineno=-1, colno=-1, origin='', value='3f1')"


def test_consume_equal_float():
    f = lambda s: fp.SequenceIterator(iter(((3, 4, s),)))
    assert EqualFloat.consume(f("a = 3.1"), None) == EqualFloat("a", 3.1).set_line_col(
        3, 4
    )
    assert EqualFloat.consume(f("a"), None) is None

    assert EqualFloat.consume(f("%a = 3.1"), None) == NotAValidIdentifier(
        "%a"
    ).set_line_col(3, 4)
    assert EqualFloat.consume(f("a = 3f1"), None) == CannotParseToFloat(
        "3f1"
    ).set_line_col(3, 4)


@pytest.mark.parametrize("klass", (MyRoot, MyRoot2))
def test_stream_block(klass):
    lines = "# hola\nx=1.0".split("\n")
    si = fp.SequenceIterator.from_lines(lines)

    mb = klass.consume(si, None)
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors


@pytest.mark.parametrize("klass", (MyRoot, MyRoot2))
def test_stream_block_error(klass):
    lines = "# hola\nx=1f0".split("\n")
    si = fp.SequenceIterator.from_lines(lines)

    mb = klass.consume(si, None)
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_line_col(0, 0),
        CannotParseToFloat("1f0").set_line_col(1, 0),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert mb.has_errors
    assert mb.errors == (CannotParseToFloat("1f0").set_line_col(1, 0),)


@pytest.mark.parametrize("klass", (MyBlock, MyBlock2))
def test_block(klass):
    lines = "@begin\n# hola\nx=1.0\n@end".split("\n")
    si = fp.SequenceIterator.from_lines(lines)

    mb = klass.consume(si, None)
    assert mb.opening == Open().set_line_col(0, 0)
    assert mb.closing == Close().set_line_col(3, 0)
    body = tuple(mb.body)
    assert len(body) == 2
    assert mb.body == (
        Comment("# hola").set_line_col(1, 0),
        EqualFloat("x", 1.0).set_line_col(2, 0),
    )

    assert tuple(mb) == (mb.opening, *mb.body, mb.closing)
    assert not mb.has_errors


@pytest.mark.parametrize("klass", (MyBlock, MyBlock2))
def test_unfinished_block(klass):
    lines = "@begin\n# hola\nx=1.0".split("\n")
    si = fp.SequenceIterator.from_lines(lines)

    mb = klass.consume(si, None)
    assert mb.opening == Open().set_line_col(0, 0)
    assert mb.closing == fp.UnexpectedEOF().set_line_col(-1, -1)
    body = tuple(mb.body)
    assert len(body) == 2
    assert mb.body == (
        Comment("# hola").set_line_col(1, 0),
        EqualFloat("x", 1.0).set_line_col(2, 0),
    )

    assert tuple(mb) == (mb.opening, *mb.body, mb.closing)
    assert mb.has_errors


def test_not_proper_statement():
    class MySt(fp.ParsedStatement):
        pass

    with pytest.raises(NotImplementedError):
        MySt.from_string("a = 1")

    with pytest.raises(NotImplementedError):
        MySt.from_string_and_config("a = 1", None)



import pathlib
from dataclasses import dataclass

import pytest

from flexparser import flexparser as fp
from flexparser.testsuite.common import (
    Close,
    Comment,
    EqualFloat,
    MyBlock,
    MyParser,
    MyRoot,
    Open,
)


def test_locator():
    this_file = pathlib.Path(__file__)

    with pytest.raises(ValueError):
        # Cannot use absolute path as target.
        assert fp.default_locator(this_file, "/temp/bla.txt")

    with pytest.raises(TypeError):
        assert fp.default_locator(str(this_file), "bla.txt")

    with pytest.raises(TypeError):
        assert fp.default_locator(str(this_file), "/temp/bla.txt")

    assert fp.default_locator(this_file, "bla.txt") == this_file.parent / "bla.txt"
    assert (
        fp.default_locator(this_file.parent, "bla.txt") == this_file.parent / "bla.txt"
    )

    with pytest.raises(ValueError):
        assert (
            fp.default_locator(this_file.parent, "../bla.txt")
            == this_file.parent / "bla.txt"
        )

    assert fp.default_locator(("pack", "nam"), "bla") == ("pack", "bla")


@pytest.mark.parametrize("definition", [MyRoot, (Comment, EqualFloat), MyParser])
def test_parse1(tmp_path, definition):
    content = "# hola\nx=1.0"
    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_text(content)

    pp = fp.parse(tmp_file, definition)

    assert len(pp) == 1

    psf = pp[list(pp.keys())[0]]
    assert not psf.has_errors
    assert psf.config is None
    assert psf.mtime == tmp_file.stat().st_mtime
    assert psf.filename == tmp_file
    assert tuple(psf.localized_errors()) == ()
    assert psf.origin == psf.filename

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors

    assert tuple(pp.iter_statements()) == (
        fp.BOS().set_line_col(0, 0),
        Comment("# hola").set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
        fp.EOS().set_line_col(-1, -1),
    )


@pytest.mark.parametrize("definition", [MyRoot, EqualFloat, MyParser])
def test_parse2(tmp_path, definition):
    content = "y = 2.0\nx=1.0"
    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_text(content)

    pp = fp.parse(tmp_file, definition)

    assert len(pp) == 1

    assert None in pp
    psf = pp[None]
    assert not psf.has_errors
    assert psf.config is None
    assert psf.mtime == tmp_file.stat().st_mtime
    assert psf.filename == tmp_file
    assert tuple(psf.localized_errors()) == ()
    assert psf.origin == psf.filename

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        EqualFloat("y", 2.0).set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors

    assert tuple(pp.iter_statements()) == (
        fp.BOS().set_line_col(0, 0),
        EqualFloat("y", 2.0).set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
        fp.EOS().set_line_col(-1, -1),
    )


@pytest.mark.parametrize(
    "definition",
    [
        MyBlock,
    ],
)
def test_parse3(tmp_path, definition):
    content = "@begin\ny = 2.0\nx=1.0\n@end"
    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_text(content)

    pp = fp.parse(tmp_file, definition)
    assert not pp.has_errors
    assert len(pp) == 1
    assert tuple(pp.localized_errors()) == ()

    assert None in pp
    psf = pp[None]
    assert not psf.has_errors
    assert psf.config is None
    assert psf.mtime == tmp_file.stat().st_mtime
    assert psf.filename == tmp_file
    assert tuple(psf.localized_errors()) == ()
    assert psf.origin == psf.filename

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 1
    mb = body[0]
    assert mb.opening == Open().set_line_col(0, 0)
    assert tuple(mb.body) == (
        EqualFloat("y", 2.0).set_line_col(1, 0),
        EqualFloat("x", 1.0).set_line_col(2, 0),
    )
    assert mb.closing == Close().set_line_col(3, 0)
    assert not mb.has_errors

    assert tuple(pp.iter_statements()) == (
        fp.BOS().set_line_col(0, 0),
        Open().set_line_col(0, 0),
        EqualFloat("y", 2.0).set_line_col(1, 0),
        EqualFloat("x", 1.0).set_line_col(2, 0),
        Close().set_line_col(3, 0),
        fp.EOS().set_line_col(-1, -1),
    )


def test_include_file(tmp_path):
    @dataclass(frozen=True)
    class Include(fp.IncludeStatement):

        value: str

        @property
        def target(self) -> str:
            return "bla2.txt"

        @classmethod
        def from_string(cls, s: str):
            if s.startswith("include"):
                return cls(s[len("include ") :].strip())

    tmp_file1 = tmp_path / "bla1.txt"
    tmp_file2 = tmp_path / "bla2.txt"
    tmp_file1.write_text("include bla2.txt\n# chau")
    tmp_file2.write_text("# hola\nx=1.0")

    pp = fp.parse(tmp_file1, (Include, Comment, EqualFloat))

    assert None in pp
    assert (tmp_file1, "bla2.txt") in pp

    assert len(pp) == 2

    assert tuple(pp.iter_statements()) == (
        fp.BOS().set_line_col(0, 0),
        fp.BOS().set_line_col(0, 0),
        Comment("# hola").set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
        fp.EOS().set_line_col(-1, -1),
        Comment("# chau").set_line_col(1, 0),
        fp.EOS().set_line_col(-1, -1),
    )


def test_resources(tmp_path):
    @dataclass(frozen=True)
    class Include(fp.IncludeStatement):

        value: str

        @property
        def target(self) -> str:
            return "bla2.txt"

        @classmethod
        def from_string(cls, s: str):
            if s.startswith("include"):
                return cls(s[len("include ") :].strip())

    pp = fp.parse(("flexparser.testsuite", "bla1.txt"), (Include, Comment, EqualFloat))

    assert len(pp) == 2

    assert tuple(pp.iter_statements()) == (
        fp.BOS().set_line_col(0, 0),
        fp.BOS().set_line_col(0, 0),
        Comment("# hola").set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
        fp.EOS().set_line_col(-1, -1),
        Comment("# chau").set_line_col(1, 0),
        fp.EOS().set_line_col(-1, -1),
    )



"""
    flexparser
    ~~~~~~~~~

    Classes for persistent caching and invalidating cached objects,
    which are built from a source object and a (potentially expensive)
    conversion function.

    :copyright: 2022 by flexparser Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


import pkg_resources

try:  # pragma: no cover
    __version__ = pkg_resources.get_distribution("flexparser").version
except Exception:  # pragma: no cover
    # we seem to have a local copy not installed without setuptools
    # so the reported version will be unknown
    __version__ = "unknown"


from .flexparser import (
    BaseIterator,
    Block,
    DelimiterMode,
    HashIterator,
    HashSequenceIterator,
    IncludeStatement,
    Multi,
    ParsedStatement,
    Parser,
    ParsingError,
    RootBlock,
    SequenceIterator,
    Single,
    StatementIterator,
    UnexpectedEOF,
    UnknownStatement,
    parse,
)

__all__ = (
    __version__,
    BaseIterator,
    Block,
    DelimiterMode,
    HashIterator,
    HashSequenceIterator,
    IncludeStatement,
    Multi,
    ParsedStatement,
    Parser,
    ParsingError,
    RootBlock,
    SequenceIterator,
    Single,
    StatementIterator,
    UnexpectedEOF,
    UnknownStatement,
    parse,
)



from flexparser import flexparser as fp
from flexparser.testsuite.common import (
    Comment,
    EqualFloat,
    MyBlock,
    MyParser,
    MyParserWithBlock,
)


def test_consume_err():
    myparser = MyParser(None)

    lines = "# hola\nx<>1.0".split("\n")
    si = fp.SequenceIterator.from_lines(lines)
    pf = myparser.consume(si)
    assert isinstance(pf.opening, fp.BOS)
    assert isinstance(pf.closing, fp.EOS)
    body = tuple(pf.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_line_col(0, 0),
        fp.UnknownStatement.from_line_col_statement(1, 0, "x<>1.0"),
    )
    assert tuple(pf) == (pf.opening, *body, pf.closing)
    assert pf.has_errors

    assert str(body[-1]) == "Could not parse 'x<>1.0' (line: 1, col: 0)"
    assert (
        str(body[-1].copy_with("blip.txt"))
        == "Could not parse 'x<>1.0' in blip.txt (line: 1, col: 0)"
    )


def test_consume():
    myparser = MyParser(None)

    lines = "# hola\nx=1.0".split("\n")
    si = fp.SequenceIterator.from_lines(lines)
    pf = myparser.consume(si)
    assert isinstance(pf.opening, fp.BOS)
    assert isinstance(pf.closing, fp.EOS)
    body = tuple(pf.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
    )
    assert tuple(pf) == (pf.opening, *body, pf.closing)
    assert not pf.has_errors


def test_parse(tmp_path):
    content = "# hola\nx=1.0"
    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_text(content)
    myparser = MyParser(None)

    psf = myparser.parse(tmp_file)
    assert not psf.has_errors
    assert psf.config is None
    assert psf.mtime == tmp_file.stat().st_mtime
    assert psf.filename == tmp_file
    assert tuple(psf.localized_errors()) == ()
    assert psf.origin == psf.filename

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors


def test_unfinished_block(tmp_path):
    content = "@begin\n# hola\nx=1.0"

    tmp_file = tmp_path / "bla.txt"
    tmp_file.write_text(content)
    myparser = MyParserWithBlock(None)

    psf = myparser.parse(tmp_file)
    assert psf.has_errors
    assert psf.config is None
    assert psf.mtime == tmp_file.stat().st_mtime
    assert psf.filename == tmp_file
    assert tuple(psf.localized_errors()) == (
        fp.UnexpectedEOF().set_line_col(-1, -1).copy_with(tmp_file),
    )
    assert psf.origin == psf.filename
    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 1
    assert isinstance(body[0], MyBlock)
    assert body[0].closing == fp.UnexpectedEOF().set_line_col(-1, -1)
    assert body[0].body == (
        Comment("# hola").set_line_col(1, 0),
        EqualFloat("x", 1.0).set_line_col(2, 0),
    )


def test_parse_resource():
    myparser = MyParser(None)

    psf = myparser.parse_resource("flexparser.testsuite", "bla2.txt")
    assert not psf.has_errors
    assert psf.config is None
    assert tuple(psf.localized_errors()) == ()
    assert psf.origin == ("flexparser.testsuite", "bla2.txt")

    # TODO:
    # assert psf.content_hash == hashlib.sha1(content.encode("utf-8")).hexdigest()

    mb = psf.parsed_source
    assert isinstance(mb.opening, fp.BOS)
    assert isinstance(mb.closing, fp.EOS)
    body = tuple(mb.body)
    assert len(body) == 2
    assert body == (
        Comment("# hola").set_line_col(0, 0),
        EqualFloat("x", 1.0).set_line_col(1, 0),
    )
    assert tuple(mb) == (mb.opening, *body, mb.closing)
    assert not mb.has_errors



import hashlib
import pickle

import pytest

import flexparser.flexparser as fp

_TEST = [
    "testing",
    "# this is a comment",
    "  # this is also a comment",
    "123",
    "456 # this is a comment",
]


def test_base_iterator():
    bi = fp.BaseIterator(iter(("spam", "ham")))
    assert bi.peek() == "spam"
    assert next(bi) == "spam"
    assert bi.peek() == "ham"
    assert next(bi) == "ham"
    with pytest.raises(StopIteration):
        bi.peek()
    with pytest.raises(StopIteration):
        next(bi)


def test_statement_iterator_default():
    si = fp.StatementIterator.from_line("spam")
    assert si.peek() == (0, "spam")
    assert tuple(si) == ((0, "spam"),)

    with pytest.raises(Exception):
        fp.StatementIterator.from_line("spam").delimiter_pattern()


def test_statement_iterator_subclass():
    class S1a(fp.StatementIterator):
        _strip_spaces = False
        _delimiters = ("#", "!")

    S1b = fp.StatementIterator.subclass_with(strip_spaces=False, delimiters=("#", "!"))
    assert S1a is not S1b
    assert S1a._strip_spaces == S1b._strip_spaces
    assert S1b._delimiters == S1b._delimiters


def test_statement_iterator_strip_spaces():
    NewStatementIterator = fp.StatementIterator.subclass_with(strip_spaces=True)

    assert tuple(NewStatementIterator.from_line("spam ")) == ((0, "spam"),)
    assert tuple(NewStatementIterator.from_line(" spam")) == ((0, "spam"),)
    assert tuple(NewStatementIterator.from_line(" spam ")) == ((0, "spam"),)

    NewStatementIterator = fp.StatementIterator.subclass_with(strip_spaces=False)

    assert tuple(NewStatementIterator.from_line("spam ")) == ((0, "spam "),)
    assert tuple(NewStatementIterator.from_line(" spam")) == ((0, " spam"),)
    assert tuple(NewStatementIterator.from_line(" spam ")) == ((0, " spam "),)

    dlm = {
        "#": (fp.DelimiterMode.SKIP, False),
    }
    NewStatementIterator = fp.StatementIterator.subclass_with(
        strip_spaces=False, delimiters=dlm
    )

    assert tuple(NewStatementIterator.from_line("spam# ham")) == (
        (0, "spam"),
        (5, " ham"),
    )
    assert tuple(NewStatementIterator.from_line("spam #ham")) == (
        (0, "spam "),
        (6, "ham"),
    )


def test_statement_iterator_single_splitter():
    dlm = {
        "#": (fp.DelimiterMode.SKIP, False),
    }
    NewStatementIterator = fp.StatementIterator.subclass_with(delimiters=dlm)

    assert tuple(NewStatementIterator.from_line("spam")) == ((0, "spam"),)
    assert tuple(NewStatementIterator.from_line("spam#ham")) == (
        (0, "spam"),
        (5, "ham"),
    )


def test_statement_iterator_multiple_splitter():
    dlm = {
        "#": (fp.DelimiterMode.SKIP, False),
        "!": (fp.DelimiterMode.SKIP, False),
    }
    NewStatementIterator = fp.StatementIterator.subclass_with(delimiters=dlm)

    assert tuple(NewStatementIterator.from_line("spam")) == ((0, "spam"),)
    assert tuple(NewStatementIterator.from_line("spam#ham!cheese")) == (
        (0, "spam"),
        (5, "ham"),
        (9, "cheese"),
    )
    assert tuple(NewStatementIterator.from_line("spam!ham#cheese")) == (
        (0, "spam"),
        (5, "ham"),
        (9, "cheese"),
    )


def test_sequence_iterator():
    si = fp.SequenceIterator.from_lines("spam \n ham".split("\n"))
    assert si.peek() == (0, 0, "spam")
    assert tuple(si) == ((0, 0, "spam"), (1, 0, "ham"))


def test_hash_sequence_iterator():
    content = "spam \n ham"
    si = fp.SequenceIterator.from_lines(content.split("\n"))
    hsi = fp.HashSequenceIterator(si)
    out = ((0, 0, "spam"), (1, 0, "ham"))
    assert tuple(hsi) == ((0, 0, "spam"), (1, 0, "ham"))
    hasher = hashlib.sha1()
    for o in out:
        hasher.update(pickle.dumps(o))
    assert hsi.hexdigest() == hasher.hexdigest()



"""
    flexparser.flexparser
    ~~~~~~~~~~~~~~~~~~~~~

    There are three type of

    - statements: single lines handled by the Definition.from_string method.
    - line directives
    - block directives:


    :copyright: 2022 by flexparser Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import collections
import dataclasses
import enum
import functools
import hashlib
import inspect
import logging
import pathlib
import pickle
import re
import typing as ty
from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from functools import cached_property
from importlib import resources
from typing import Dict, Optional, Tuple, Type

_LOGGER = logging.getLogger("flexparser")

_SENTINEL = object()


################
# Exceptions
################


@dataclass(frozen=True)
class Element:
    """Base class for elements within a source file to be parsed."""

    lineno: int = dataclasses.field(init=False, default=-1)
    colno: int = dataclasses.field(init=False, default=-1)

    @property
    def format_line_col(self):
        return f"(line: {self.lineno}, col: {self.colno})"

    def set_line_col(self, lineno, colno):
        object.__setattr__(self, "lineno", lineno)
        object.__setattr__(self, "colno", colno)
        return self


@dataclass(frozen=True)
class ParsingError(Element, Exception):
    """Base class for all exceptions in this package."""

    origin: ty.Union[str, ty.Tuple[str, str], pathlib.Path] = dataclasses.field(
        init=False, default=""
    )

    @property
    def origin_(self) -> str:
        if isinstance(self.origin, tuple):
            return f"resource (package: {self.origin[0]}, name: {self.origin[1]})"
        return str(self.origin)

    def copy_with(self, origin):
        d = dataclasses.asdict(self)
        d.pop("lineno")
        d.pop("colno")
        d.pop("origin")
        obj = self.__class__(**d)
        obj.set_line_col(self.lineno, self.colno)
        object.__setattr__(obj, "origin", origin)
        return obj

    def __str__(self):
        return Element.__str__(self)


@dataclass(frozen=True)
class UnknownStatement(ParsingError):
    """A string statement could not bee parsed."""

    statement: str

    def __str__(self):
        if self.origin:
            return f"Could not parse '{self.statement}' in {self.origin_} (line: {self.lineno}, col: {self.colno})"
        return f"Could not parse '{self.statement}' (line: {self.lineno}, col: {self.colno})"

    @classmethod
    def from_line_col_statement(cls, lineno, colno, statement):
        obj = cls(statement)
        obj.set_line_col(lineno, colno)
        return obj


@dataclass(frozen=True)
class UnexpectedEOF(ParsingError):
    """End of file was found within an open block."""


#################
# Useful methods
#################


def _yield_types(
    obj, valid_subclasses=(object,), recurse_origin=(tuple, list, ty.Union)
):
    """Recursively transverse type annotation if the
    origin is any of the types in `recurse_origin`
    and yield those type which are subclasses of `valid_subclasses`.

    """
    if ty.get_origin(obj) in recurse_origin:
        for el in ty.get_args(obj):
            yield from _yield_types(el, valid_subclasses, recurse_origin)
    else:
        if inspect.isclass(obj) and issubclass(obj, valid_subclasses):
            yield obj


class classproperty:  # noqa N801
    """Decorator for a class property

    In Python 3.9+ can be replaced by

        @classmethod
        @property
        def myprop(self):
            return 42

    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


def is_empty_pattern(pattern: str | re.Pattern) -> bool:
    """True if the regex pattern string is empty
    or the compiled version comes from an empty string.
    """
    if isinstance(pattern, str):
        return not bool(pattern)
    return not bool(getattr(pattern, "pattern", ""))


def isplit(
    pattern: str | re.Pattern, line: str
) -> Iterator[tuple[int, str, Optional[str]]]:
    """Yield position, strings between matches and match group (delimiter)
    in a regex pattern applied to a line.

    Returns None as delimiter for last element.
    """
    col = 0
    if not is_empty_pattern(pattern):
        for m in re.finditer(pattern, line):
            part = line[col : m.start()]
            yield col, part, m.group(0)
            col = m.end()

    yield col, line[col:], None


class DelimiterMode(enum.Enum):
    """Specifies how to deal with delimiters while parsing."""

    #: Skip delimiter in output string
    SKIP = 0

    #: Keep delimiter with previous string.
    WITH_PREVIOUS = 1

    #: Keep delimiter with next string.
    WITH_NEXT = 2


@functools.lru_cache
def _build_delimiter_pattern(pattern_tuple: Tuple[str, ...]) -> re.Pattern:
    """Compile a tuple of delimiters into a regex expression with a capture group
    around the delimiter.
    """
    return re.compile("|".join(f"({re.escape(el)})" for el in pattern_tuple))


def isplit_mode(
    delimiters: Dict[str, Tuple[DelimiterMode, bool]], line: str
) -> Iterator[tuple[int, str]]:
    """Yield position and strings between delimiters.

    The delimiters are specified with the keys of the delimiters dict.
    The dict files can be used to further customize the iterator. Each
    consist of a tuple of two elements:
      1. A value of the DelimiterMode to indicate what to do with the
         delimiter string: skip it, attach keep it with previous or next string
      2. A boolean indicating if parsing should stop after fiSBT
         encountering this delimiter.

    Empty strings are not yielded.
    """
    pattern = _build_delimiter_pattern(tuple(delimiters.keys()))
    dragged = ""
    break_next = False
    for col, part, dlm in isplit(pattern, line):
        if break_next:
            yield col - len(dragged), dragged + line[col:]
            break

        next_drag = ""
        if dlm is None:
            delimiter_mode, break_next = DelimiterMode.SKIP, False
        else:
            delimiter_mode, break_next = delimiters[dlm]

            if delimiter_mode == DelimiterMode.WITH_PREVIOUS:
                part = part + dlm
            elif delimiter_mode == DelimiterMode.WITH_NEXT:
                next_drag = dlm

        if part:
            yield col - len(dragged), dragged + part

        dragged = next_drag


############
# Iterators
############

T = ty.TypeVar("T")


class BaseIterator(ty.Generic[T], Iterator[T]):
    """Base class for iterator that provides the ability to peek."""

    _cache: ty.Deque[T]

    def __init__(self, iterator: Iterator[T]):
        self._it = iterator
        self._cache = collections.deque()

    def __iter__(self):
        return self

    def peek(self, default=_SENTINEL) -> T:
        """Return the item that will be next returned from ``next()``.

        Return ``default`` if there are no items left. If ``default`` is not
        provided, raise ``StopIteration``.

        """
        if not self._cache:
            try:
                self._cache.append(next(self._it))
            except StopIteration:
                if default is _SENTINEL:
                    raise
                return default
        return self._cache[0]

    def __next__(self) -> T:
        if self._cache:
            return self._cache.popleft()

        return next(self._it)


class StatementIterator(BaseIterator[Tuple[int, str]]):
    """Yield position and statement within a string,
    ending when all the line is consumed.

    Elements are separated by delimiter/s and leading and trailing spaces
    are (optionally) removed.

    By default, spaces are removed and no delimiter is defined.

    >>> si = StatementIterator(" spam ")
    >>> si.peek()
    0, "spam"
    >>> tuple(si)
    ((0, "spam"), )

    The behavior can be customized by subclassing

    >>> class CustomStatementIterator(StatementIterator):
    ...     _strip_spaces = False
    ...     _delimiters = {"#": (DelimiterMode.WITH_NEXT, True)}

    or equivalent:

    >>> CustomStatementIterator = StatementIterator.subclass_with(strip_spaces=False,
    ...                                                           delimiters = {"#": (DelimiterMode.WITH_NEXT, True)})

    """

    _strip_spaces: bool = True
    _delimiters: Optional[Dict[str, Tuple[DelimiterMode, bool]]] = None

    @classmethod
    def from_line(cls, line: str):
        if cls._delimiters:
            it = iter(isplit_mode(cls._delimiters, line))
        else:
            it = iter(((0, line),))

        if cls._strip_spaces:
            it = ((colno, s.strip()) for colno, s in it)

        return cls(it)

    @classmethod
    def subclass_with(cls, *, strip_spaces=True, delimiters: dict = None):
        """Creates a new subclass in one line."""

        return type(
            "CustomStatementIterator",
            (cls,),
            dict(_strip_spaces=strip_spaces, _delimiters=delimiters or {}),
        )


class SequenceIterator(BaseIterator[Tuple[int, int, str]]):
    """Yield line, position and statement within elements of a sequence.

    These sub-elements are parsed from the elements used a StatementIterator,
    and it can be customized by subclassing this class.

    >>> CustomStatementIterator = StatementIterator.subclass_with(strip_spaces=False)
    >>> class CustomSequenceIterator(SequenceIterator):
    ...     _statement_iterator_class = CustomStatementIterator
    """

    _statement_iterator_class: Type[StatementIterator] = StatementIterator

    @classmethod
    def from_lines(cls, lines: Iterable[str]):
        it = (
            (lineno, colno, s)
            for lineno, line in enumerate(lines)
            for (colno, s) in cls._statement_iterator_class.from_line(line)
            if s
        )

        return cls(it)

    @classmethod
    def subclass_with(cls, *, statement_iterator_class):
        """Creates a new subclass in one line."""

        return type(
            "CustomSequenceIterator",
            (cls,),
            dict(_statement_iterator_class=statement_iterator_class),
        )


class HashIterator(BaseIterator[T]):
    """Iterates through an interator while hashing the content."""

    def __init__(self, iterator: Iterator[T]):
        super().__init__(iterator)
        self._hasher = hashlib.sha1()

    def __iter__(self):
        return self

    def __next__(self):
        el = super().__next__()
        self._hasher.update(pickle.dumps(el))
        return el

    def hexdigest(self):
        return self._hasher.hexdigest()


class HashSequenceIterator(HashIterator, SequenceIterator):
    pass


###########
# Parsing
###########

# Configuration type
CT = ty.TypeVar("CT")
PST = ty.TypeVar("PST", bound="ParsedStatement")
LineColStr = Tuple[int, int, str]
FromString = ty.Union[None, PST, ParsingError]
Consume = ty.Union[PST, ParsingError]
NullableConsume = ty.Union[None, PST, ParsingError]

Single = ty.Union[PST, ParsingError]
Multi = ty.Tuple[ty.Union[PST, ParsingError], ...]


@dataclass(frozen=True)
class ParsedStatement(ty.Generic[CT], Element):
    """A single parsed statement.

    In order to write your own, you need to subclass it as a
    frozen dataclass and implement the parsing logic by overriding
    `from_string` classmethod.

    Takes two arguments: the string to parse and an object given
    by the parser which can be used to store configuration information.

    It should return an instance of this class if parsing
    was successful or None otherwise
    """

    @classmethod
    def from_string(cls: Type[PST], s: str) -> FromString[PST]:
        """Parse a string into a ParsedStatement.

        Return files and their meaning:
        1. None: the string cannot be parsed with this class.
        2. A subclass of ParsedStatement: the string was parsed successfully
        3. A subclass of ParsingError the string could be parsed with this class but there is
           an error.
        """
        raise NotImplementedError(
            "ParsedStatement subclasses must implement "
            "'from_string' or 'from_string_and_config'"
        )

    @classmethod
    def from_string_and_config(cls: Type[PST], s: str, config: CT) -> FromString[PST]:
        """Parse a string into a ParsedStatement.

        Return files and their meaning:
        1. None: the string cannot be parsed with this class.
        2. A subclass of ParsedStatement: the string was parsed successfully
        3. A subclass of ParsingError the string could be parsed with this class but there is
           an error.
        """
        return cls.from_string(s)

    @classmethod
    def consume(
        cls: Type[PST], sequence_iterator: SequenceIterator, config: CT
    ) -> NullableConsume[PST]:
        """Peek into the iterator and try to parse.

        Return files and their meaning:
        1. None: the string cannot be parsed with this class, the iterator is kept an the current place.
        2. a subclass of ParsedStatement: the string was parsed successfully, advance the iterator.
        3. a subclass of ParsingError: the string could be parsed with this class but there is
           an error, advance the iterator.
        """
        lineno, colno, statement = sequence_iterator.peek()
        parsed_statement = cls.from_string_and_config(statement, config)
        if parsed_statement is None:
            return None
        next(sequence_iterator)
        parsed_statement.set_line_col(lineno, colno)
        return parsed_statement


OPST = ty.TypeVar("OPST", bound="ParsedStatement")
IPST = ty.TypeVar("IPST", bound="ParsedStatement")
CPST = ty.TypeVar("CPST", bound="ParsedStatement")
BT = ty.TypeVar("BT", bound="Block")
RBT = ty.TypeVar("RBT", bound="RootBlock")


@dataclass(frozen=True)
class Block(ty.Generic[OPST, IPST, CPST, CT]):
    """A sequence of statements with an opening, body and closing."""

    opening: Consume[OPST]
    body: Tuple[Consume[IPST], ...]
    closing: Consume[CPST]

    @classmethod
    def subclass_with(cls, *, opening=None, body=None, closing=None):
        @dataclass(frozen=True)
        class CustomBlock(Block):
            pass

        if opening:
            CustomBlock.__annotations__["opening"] = Single[ty.Union[opening]]
        if body:
            CustomBlock.__annotations__["body"] = Multi[ty.Union[body]]
        if closing:
            CustomBlock.__annotations__["closing"] = Single[ty.Union[closing]]

        return CustomBlock

    def __iter__(self) -> Iterator[Element]:
        yield self.opening
        for el in self.body:
            if isinstance(el, Block):
                yield from el
            else:
                yield el
        yield self.closing

    ###################################################
    # Convenience methods to iterate parsed statements
    ###################################################

    _ElementT = ty.TypeVar("_ElementT", bound=Element)

    def filter_by(self, *klass: Type[_ElementT]) -> Iterator[_ElementT]:
        """Yield elements of a given class or classes."""
        yield from (el for el in self if isinstance(el, klass))  # noqa Bug in pycharm.

    @cached_property
    def errors(self) -> ty.Tuple[ParsingError, ...]:
        """Tuple of errors found."""
        return tuple(self.filter_by(ParsingError))

    @property
    def has_errors(self) -> bool:
        """True if errors were found during parsing."""
        return bool(self.errors)

    ####################
    # Statement classes
    ####################

    @classproperty
    def opening_classes(cls) -> Iterator[Type[OPST]]:
        """Classes representing any of the parsed statement that can open this block."""
        opening = ty.get_type_hints(cls)["opening"]
        yield from _yield_types(opening, ParsedStatement)

    @classproperty
    def body_classes(cls) -> Iterator[Type[IPST]]:
        """Classes representing any of the parsed statement that can be in the body."""
        body = ty.get_type_hints(cls)["body"]
        yield from _yield_types(body, (ParsedStatement, Block))

    @classproperty
    def closing_classes(cls) -> Iterator[Type[CPST]]:
        """Classes representing any of the parsed statement that can close this block."""
        closing = ty.get_type_hints(cls)["closing"]
        yield from _yield_types(closing, ParsedStatement)

    ##########
    # Consume
    ##########

    @classmethod
    def consume_opening(
        cls: Type[BT], sequence_iterator: SequenceIterator, config: CT
    ) -> NullableConsume[OPST]:
        """Peek into the iterator and try to parse with any of the opening classes.

        See `ParsedStatement.consume` for more details.
        """
        for c in cls.opening_classes:
            el = c.consume(sequence_iterator, config)
            if el is not None:
                return el
        return None

    @classmethod
    def consume_body(
        cls, sequence_iterator: SequenceIterator, config: CT
    ) -> Consume[IPST]:
        """Peek into the iterator and try to parse with any of the body classes.

        If the statement cannot be parsed, a UnknownStatement is returned.
        """
        for c in cls.body_classes:
            el = c.consume(sequence_iterator, config)
            if el is not None:
                return el
        el = next(sequence_iterator)
        return UnknownStatement.from_line_col_statement(*el)

    @classmethod
    def consume_closing(
        cls: Type[BT], sequence_iterator: SequenceIterator, config: CT
    ) -> NullableConsume[CPST]:
        """Peek into the iterator and try to parse with any of the opening classes.

        See `ParsedStatement.consume` for more details.
        """
        for c in cls.closing_classes:
            el = c.consume(sequence_iterator, config)
            if el is not None:
                return el
        return None

    @classmethod
    def consume(
        cls: Type[BT], sequence_iterator: SequenceIterator, config: CT
    ) -> Optional[BT]:
        """Try consume the block.

        Possible outcomes:
        1. The opening was not matched, return None.
        2. A subclass of Block, where body and closing migh contain errors.
        """
        opening = cls.consume_opening(sequence_iterator, config)
        if opening is None:
            return None
        body = []
        closing = None
        while closing is None:
            try:
                closing = cls.consume_closing(sequence_iterator, config)
                if closing is not None:
                    continue
                el = cls.consume_body(sequence_iterator, config)
                body.append(el)
            except StopIteration:
                closing = cls.on_stop_iteration(config)

        return cls(opening, tuple(body), closing)

    @classmethod
    def on_stop_iteration(cls, config):
        return UnexpectedEOF()


class BOS(ParsedStatement[CT]):
    """Beginning of sequence."""

    @classmethod
    def from_string_and_config(cls: Type[PST], s: str, config: CT) -> FromString[PST]:
        return cls()


class EOS(ParsedStatement[CT]):
    """End of sequence."""

    @classmethod
    def from_string_and_config(cls: Type[PST], s: str, config: CT) -> FromString[PST]:
        return cls()


class RootBlock(ty.Generic[IPST, CT], Block[BOS, IPST, EOS, CT]):
    """A sequence of statement flanked by the beginning and ending of stream."""

    opening: Single[BOS]
    closing: Single[EOS]

    @classmethod
    def subclass_with(cls, *, body=None):
        @dataclass(frozen=True)
        class CustomRootBlock(RootBlock):
            pass

        if body:
            CustomRootBlock.__annotations__["body"] = Multi[ty.Union[body]]

        return CustomRootBlock

    @classmethod
    def consume_opening(
        cls: Type[RBT], sequence_iterator: SequenceIterator, config: CT
    ) -> NullableConsume[BOS]:
        return BOS().set_line_col(0, 0)

    @classmethod
    def consume(cls: Type[RBT], sequence_iterator: SequenceIterator, config: CT) -> RBT:
        block = super().consume(sequence_iterator, config)
        if block is None:
            raise ValueError(
                "Implementation error, 'RootBlock.consume' should never return None"
            )
        return block

    @classmethod
    def consume_closing(
        cls: Type[RBT], sequence_iterator: SequenceIterator, config: CT
    ) -> NullableConsume[EOS]:
        return None

    @classmethod
    def on_stop_iteration(cls, config):
        return EOS()


#################
# Source parsing
#################

ResourceT = ty.Tuple[str, str]  # package name, resource name
StrictLocationT = ty.Union[pathlib.Path, ResourceT]
SourceLocationT = ty.Union[str, StrictLocationT]


@dataclass(frozen=True)
class _ParsedCommon(ty.Generic[RBT, CT]):

    parsed_source: RBT

    # SHA-1 hash.
    content_hash: str

    # Parser configuration.
    config: CT

    @property
    def origin(self) -> StrictLocationT:
        raise NotImplementedError

    @cached_property
    def has_errors(self) -> bool:
        return self.parsed_source.has_errors

    def localized_errors(self):
        for err in self.parsed_source.errors:
            yield err.copy_with(self.origin)


@dataclass(frozen=True)
class ParsedSourceFile(_ParsedCommon[RBT, CT], ty.Generic[RBT, CT]):
    """The parsed representation of a file."""

    # Fullpath of the original file.
    filename: pathlib.Path

    # Modification time of the file.
    mtime: float

    @property
    def origin(self) -> pathlib.Path:
        return self.filename


@dataclass(frozen=True)
class ParsedResource(_ParsedCommon[RBT, CT], ty.Generic[RBT, CT]):
    """The parsed representation of a python resource."""

    # Fullpath of the original file, None if a text was provided
    package: str
    resource_name: str

    @property
    def origin(self) -> ResourceT:
        return self.package, self.resource_name


@dataclass(frozen=True)
class CannotParseResourceAsFile(Exception):
    """The requested python package resource cannot be located as a file
    in the file system.
    """

    package: str
    resource_name: str


class Parser(ty.Generic[RBT, CT]):
    """Parser class."""

    #: class to iterate through statements in a source unit.
    _sequence_iterator_class: Type[SequenceIterator] = SequenceIterator

    #: root block class containing statements and blocks can be parsed.
    _root_block_class: Type[RBT]

    #: source file text encoding.
    _encoding = "utf-8"

    #: configuration passed to from_string functions.
    _config: CT

    #: try to open resources as files.
    _prefer_resource_as_file: bool

    def __init__(self, config: CT, prefer_resource_as_file=True):
        self._config = config
        self._prefer_resource_as_file = prefer_resource_as_file

    def consume(self, sequence_iterator: SequenceIterator) -> RBT:
        return self._root_block_class.consume(sequence_iterator, self._config)

    def parse(
        self, source_location: SourceLocationT
    ) -> ty.Union[ParsedSourceFile[RBT, CT], ParsedResource[RBT, CT]]:
        """Parse a file into a ParsedSourceFile or ParsedResource.

        Parameters
        ----------
        source_location:
            if str or pathlib.Path is interpreted as a file.
            if (str, str) is interpreted as (package, resource) using the resource python api.
        """
        if isinstance(source_location, tuple) and len(source_location) == 2:
            if self._prefer_resource_as_file:
                try:
                    return self.parse_resource_from_file(*source_location)
                except CannotParseResourceAsFile:
                    pass
            return self.parse_resource(*source_location)

        if isinstance(source_location, str):
            return self.parse_file(pathlib.Path(source_location))

        if isinstance(source_location, pathlib.Path):
            return self.parse_file(source_location)

        raise TypeError(
            f"Unknown type {type(source_location)}, "
            "use str or pathlib.Path for files or "
            "(package: str, resource_name: str) tuple "
            "for a resource."
        )

    def parse_file(self, path: pathlib.Path) -> ParsedSourceFile[RBT, CT]:
        """Parse a file into a ParsedSourceFile.

        Parameters
        ----------
        path
            path of the file.
        """
        with path.open(mode="r", encoding=self._encoding) as fi:
            sic = self._sequence_iterator_class.from_lines(
                map(lambda s: s.strip("\r\n"), fi)
            )
            hsi = HashSequenceIterator(sic)
            body = self.consume(hsi)
            return ParsedSourceFile(
                body,
                hsi.hexdigest(),
                self._config,
                path,
                path.stat().st_mtime,
            )

    def parse_resource_from_file(
        self, package: str, resource_name: str
    ) -> ParsedSourceFile[RBT, CT]:
        """Parse a resource into a ParsedSourceFile, opening as a file.

        Parameters
        ----------
        package
            package name where the resource is located.
        resource_name
            name of the resource
        """
        with resources.path(package, resource_name) as p:
            path = p.resolve()

        if path.exists():
            return self.parse_file(path)

        raise CannotParseResourceAsFile(package, resource_name)

    def parse_resource(
        self, package: str, resource_name: str
    ) -> ParsedResource[RBT, CT]:
        """Parse a resource into a ParsedResource.

        Parameters
        ----------
        package
            package name where the resource is located.
        resource_name
            name of the resource
        """
        with resources.open_text(package, resource_name, encoding=self._encoding) as fi:
            sic = self._sequence_iterator_class.from_lines(
                map(lambda s: s.strip("\r\n"), fi)
            )
            hsi = HashSequenceIterator(sic)
            body = self.consume(hsi)
        return ParsedResource(
            body,
            hsi.hexdigest(),
            self._config,
            package,
            resource_name,
        )


##########
# Project
##########


class IncludeStatement(ParsedStatement):
    """ "Include statements allow to merge files."""

    @property
    def target(self) -> str:
        raise NotImplementedError(
            "IncludeStatement subclasses must implement target property."
        )


class ParsedProject(
    dict[ty.Optional[StrictLocationT], ty.Union[ParsedSourceFile, ParsedResource]]
):
    """Collection of files, independent or connected via IncludeStatement.

    Keys are either an absolute pathname  or a tuple package name, resource name.

    None is the name of the root.

    """

    @cached_property
    def has_errors(self) -> bool:
        return any(el.has_errors for el in self.values())

    def localized_errors(self):
        for el in self.values():
            yield from el.localized_errors()

    def _iter_statements(self, items, seen, include_only_once):
        """Iter all definitions in the order they appear,
        going into the included files.
        """
        for source_location, parsed in items:
            seen.add(source_location)
            for parsed_statement in parsed.parsed_source:
                if isinstance(parsed_statement, IncludeStatement):
                    location = parsed.origin, parsed_statement.target
                    if location in seen and include_only_once:
                        raise ValueError(f"{location} was already included.")
                    yield from self._iter_statements(
                        ((location, self[location]),), seen, include_only_once
                    )
                else:
                    yield parsed_statement

    def iter_statements(self, include_only_once=True):
        """Iter all definitions in the order they appear,
        going into the included files.

        Parameters
        ----------
        include_only_once
            if true, each file cannot be included more than once.
        """
        yield from self._iter_statements([(None, self[None])], set(), include_only_once)


def default_locator(source_location: StrictLocationT, target: str) -> StrictLocationT:
    """Return a new location from current_location and target."""

    if isinstance(source_location, pathlib.Path):
        current_location = pathlib.Path(source_location).resolve()

        if current_location.is_file():
            current_path = source_location.parent
        else:
            current_path = source_location

        target_path = pathlib.Path(target)
        if target_path.is_absolute():
            raise ValueError(
                f"Cannot refer to absolute paths in import statements ({source_location}, {target})."
            )

        tmp = (current_path / target_path).resolve()
        if not tmp.is_relative_to(current_path):
            raise ValueError(
                f"Cannot refer to locations above the current location ({source_location}, {target})"
            )

        return tmp.absolute()

    elif isinstance(source_location, tuple) and len(source_location) == 2:
        return source_location[0], target

    raise TypeError(
        f"Cannot handle type {type(source_location)}, "
        "use str or pathlib.Path for files or "
        "(package: str, resource_name: str) tuple "
        "for a resource."
    )


SpecT = ty.Union[
    ty.Type[Parser],
    ty.Union[ty.Type[Block], ty.Type[ParsedStatement]],
    ty.Iterable[ty.Union[ty.Type[Block], ty.Type[ParsedStatement]]],
    ty.Type[RootBlock],
]


def parse(
    entry_point: SourceLocationT,
    spec: SpecT,
    config=None,
    *,
    strip_spaces=True,
    delimiters=None,
    locator=default_locator,
    prefer_resource_as_file=True,
) -> ParsedProject:
    """Parse sources into a ParsedProject."""

    if isinstance(spec, type) and issubclass(spec, Parser):
        CustomParser = spec
    else:
        if isinstance(spec, (tuple, list)):

            for el in spec:
                if not issubclass(el, (Block, ParsedStatement)):
                    raise TypeError(
                        "Elements in root_block_class must be of type Block or ParsedStatement, "
                        f"not {el}"
                    )

            @dataclass(frozen=True)
            class CustomRootBlock(RootBlock):
                pass

            CustomRootBlock.__annotations__["body"] = Multi[ty.Union[spec]]

        elif isinstance(spec, type) and issubclass(spec, RootBlock):

            CustomRootBlock = spec

        elif isinstance(spec, type) and issubclass(spec, (Block, ParsedStatement)):

            @dataclass(frozen=True)
            class CustomRootBlock(RootBlock):
                pass

            CustomRootBlock.__annotations__["body"] = Multi[spec]

        else:
            raise TypeError(
                "`spec` must be of type RootBlock or tuple of type Block or ParsedStatement, "
                f"not {type(spec)}"
            )

        class CustomParser(Parser):

            _sequence_iterator_class = SequenceIterator.subclass_with(
                statement_iterator_class=StatementIterator.subclass_with(
                    strip_spaces=strip_spaces, delimiters=delimiters
                )
            )
            _root_block_class = CustomRootBlock

    parser = CustomParser(config, prefer_resource_as_file=prefer_resource_as_file)

    pp = ParsedProject()

    # : ty.List[Optional[ty.Union[LocatorT, str]], ...]
    pending: ty.List[ty.Optional[ty.Tuple[StrictLocationT, str]]] = []
    if isinstance(entry_point, (str, pathlib.Path)):
        entry_point = pathlib.Path(entry_point)
        if not entry_point.is_absolute():
            entry_point = pathlib.Path.cwd() / entry_point

    elif not (isinstance(entry_point, tuple) and len(entry_point) == 2):
        raise TypeError(
            f"Cannot handle type {type(entry_point)}, "
            "use str or pathlib.Path for files or "
            "(package: str, resource_name: str) tuple "
            "for a resource."
        )

    pp[None] = parsed = parser.parse(entry_point)
    pending.extend(
        (parsed.origin, el.target)
        for el in parsed.parsed_source.filter_by(IncludeStatement)
    )

    while pending:
        source_location, target = pending.pop(0)
        pp[(source_location, target)] = parsed = parser.parse(
            locator(source_location, target)
        )
        pending.extend(
            (parsed.origin, el.target)
            for el in parsed.parsed_source.filter_by(IncludeStatement)
        )

    return pp



import re
import typing

import flexparser.flexparser as fp


def test_yield_types():
    class X:
        pass

    assert tuple(fp._yield_types(float)) == (float,)
    assert tuple(fp._yield_types(X)) == (X,)
    assert tuple(fp._yield_types(X())) == ()


def test_yield_types_container():
    class X:
        pass

    o = typing.Tuple[float, X]
    assert tuple(fp._yield_types(o)) == (float, X)

    o = typing.Tuple[float, ...]
    assert tuple(fp._yield_types(o)) == (float,)

    o = typing.Tuple[typing.Union[float, X], ...]
    assert tuple(fp._yield_types(o)) == (float, X)


def test_yield_types_union():
    class X:
        pass

    o = typing.Union[float, X]
    assert tuple(fp._yield_types(o)) == (float, X)


def test_yield_types_list():
    o = typing.List[float]
    assert tuple(fp._yield_types(o)) == (float,)


def test_is_empty_pattern():
    assert fp.is_empty_pattern("")
    assert fp.is_empty_pattern(re.compile(""))
    assert not fp.is_empty_pattern("!")
    assert not fp.is_empty_pattern(re.compile("!"))


def test_isplit():
    assert tuple(fp.isplit("", "spam!is#ham")) == ((0, "spam!is#ham", None),)
    assert tuple(fp.isplit("#", "spamis#ham")) == ((0, "spamis", "#"), (7, "ham", None))
    assert tuple(fp.isplit("!|#", "spam!is#ham")) == (
        (0, "spam", "!"),
        (5, "is", "#"),
        (8, "ham", None),
    )
    assert tuple(fp.isplit("#", "#spamisham")) == (
        (0, "", "#"),
        (1, "spamisham", None),
    )


def test_isplit_mode():
    delimiters = {
        "!": (fp.DelimiterMode.SKIP, False),
        "#": (fp.DelimiterMode.SKIP, False),
    }
    assert tuple(fp.isplit_mode(delimiters, "spam!is#ham")) == (
        (0, "spam"),
        (5, "is"),
        (8, "ham"),
    )
    assert tuple(fp.isplit_mode(delimiters, "#spam!is#ham")) == (
        (1, "spam"),
        (6, "is"),
        (9, "ham"),
    )

    delimiters = {
        "!": (fp.DelimiterMode.WITH_NEXT, False),
        "#": (fp.DelimiterMode.WITH_NEXT, False),
    }
    assert tuple(fp.isplit_mode(delimiters, "spam!is#ham")) == (
        (0, "spam"),
        (4, "!is"),
        (7, "#ham"),
    )

    delimiters = {
        "!": (fp.DelimiterMode.WITH_PREVIOUS, False),
        "#": (fp.DelimiterMode.WITH_PREVIOUS, False),
    }
    assert tuple(fp.isplit_mode(delimiters, "spam!is#ham")) == (
        (0, "spam!"),
        (5, "is#"),
        (8, "ham"),
    )


def test_isplit_mode_break():
    delimiters = {
        "!": (fp.DelimiterMode.SKIP, True),
        "#": (fp.DelimiterMode.SKIP, True),
    }
    assert tuple(fp.isplit_mode(delimiters, "spam!is#ham")) == (
        (0, "spam"),
        (5, "is#ham"),
    )

    delimiters = {
        "!": (fp.DelimiterMode.WITH_NEXT, True),
        "#": (fp.DelimiterMode.WITH_NEXT, True),
    }
    assert tuple(fp.isplit_mode(delimiters, "spam!is#ham")) == (
        (0, "spam"),
        (4, "!is#ham"),
    )

    delimiters = {
        "!": (fp.DelimiterMode.WITH_PREVIOUS, True),
        "#": (fp.DelimiterMode.WITH_PREVIOUS, True),
    }
    assert tuple(fp.isplit_mode(delimiters, "spam!is#ham")) == (
        (0, "spam!"),
        (5, "is#ham"),
    )
